#!/usr/bin/env python

#############################################################################
# @ddblock_begin copyright 
############################################################################
# Copyright (c) 1999-2014
# Maryland DSPCAD Research Group, The University of Maryland at College Park 
#
# Permission is hereby granted, without written agreement and without
# license or royalty fees, to use, copy, modify, and distribute this
# software and its documentation for any purpose, provided that the above
# copyright notice and the following two paragraphs appear in all copies
# of this software.
# 
# IN NO EVENT SHALL THE UNIVERSITY OF MARYLAND BE LIABLE TO ANY PARTY
# FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
# ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF
# THE UNIVERSITY OF MARYLAND HAS BEEN ADVISED OF THE POSSIBILITY OF
# SUCH DAMAGE.
# 
# THE UNIVERSITY OF MARYLAND SPECIFICALLY DISCLAIMS ANY WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. THE SOFTWARE
# PROVIDED HEREUNDER IS ON AN "AS IS" BASIS, AND THE UNIVERSITY OF
# MARYLAND HAS NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES,
# ENHANCEMENTS, OR MODIFICATIONS.
############################################################################

# @ddblock_end copyright
#############################################################################


# Author: William Plishker

# The usage of \ddcode{ddblock.py} is briefly listed by running it
# with no arguments: \ddcode{ddblock} extracts or replaces block tags.
# In the case of \ddcode{replace}, \ddcode{ddblock} modifies target
# files.  \ddcode{extract} and \ddcode{replace} require an exact match
# of the block tags, while extract_prefix requires only a prefix
# match.  \ddcode{ddblock} operates on either a single file or a
# recursively on a single directory with the following usage:

# Usage 1: ./ddblock [extract,extract_prefix, replace] block_tag [more block_tags] target_directory
# Usage 2: ./ddblock [extract,extract_prefix, replace] block_tag [more block_tags] target_file

# NOTE: If used in the first mode with \ddcode{replace}, the script
# will recursively MODIFY files in the target directory changing all
# blocks to those entries found in the block directory

import sys
import os
import re

if(os.environ.has_key("UXTMP")):
    tmppath =  os.environ["UXTMP"]
else:
    print "Warning: environment variable UXTMP not set"
    tmppath =  "/tmp"
    print "         defaulting to:", tmppath


if(len(sys.argv)<4): 
    print ""
    print "Usage 1: ./ddblock [extract,extract_prefix, replace] block_tag [more block_tags] target_directory"
    print "Usage 2: ./ddblock [extract,extract_prefix, replace] block_tag [more block_tags] target_file"
    print "    extract        - extracts blocks in \"target\" with \"block_tag\"(s) labels will print to stdout"
    print "    extract_prefix - extracts blocks in \"target\" with a prefix match \"block_tag\"(s) labels will print to stdout"
    print "    replace        - ddblocks in \"target\" with \"block_tag\"(s) will be changed to "
    print "              those entries found in the block directory from $UXBLOCKPATH"
    print "              Backup copies will be saved to:", tmppath, 
#    print "            (this directory can be set through the environment"
#    print "             variable $UXBLOCKPATH - be sure to use only absolute paths)"
    exit()

#else pull the args
mode=sys.argv[1]
tags=["0"]*(len(sys.argv)-3)
for i in range(len(sys.argv)-3):
    tags[i]=sys.argv[i+2]
target=sys.argv[len(sys.argv)-1]

extract_prefix = False
if(mode[0]=="e"):
    if("p" in mode):
        extract_prefix = True
    mode = "extract"

elif(mode[0]=="r"):
    mode = "replace"
    #check if there is a custom setting for the blocks through
    #the environment variable $UXBLOCKPATH

    ## we only really need a block path if we are replacing text 
    if(os.environ.has_key("UXBLOCKPATH")):
        blockpath =  os.environ["UXBLOCKPATH"]
    else:
        print "Error: environment variable UXBLOCKPATH not set"
        exit(1)
else:
    print "ERROR: you must specify extract or replace"
    print "run ddblock with no arguments for more help"
    exit(1)

if(os.path.isdir(target)==False and os.path.isfile(target)==False):
    print "ERROR: the last argument passed to ddblock must be a file or directory"
    print "run ddblock with no arguments for more help"
    exit(1)


###################### verbosePrint  ########################
#this method is used for verbosePrinting based on the UXVERBOSE
#variable settings
def verbosePrint(item):
    if(os.environ.has_key("UXVERBOSE")):
        if(os.environ["UXVERBOSE"]!=0):
            print item


###################### extractBlock method ############################
#define the extractBlock method which takes one file looks for a specific ddblock tag
def extractBlock(tag, file, extract_prefix=False):
    #ignore some files, like starting with ".", being in a directory
    #starting with "." (like .cvs or .svn), or a file ending in "~" or
    #.bak.  To implement this the regexs are geared to match valid
    #files.
    reslashdot = re.compile("/[^\.]")
    redot      = re.compile("^[^\.]")
    retilda    = re.compile("[^\~]$")
    rebak      = re.compile("(?!\.bak).{4}$")
    if((reslashdot.search(file) or redot.search(file)) 
       and retilda.search(file) and (rebak.search(file) or len(file)<4) ):

        assert os.path.isdir(file)==False

        fp = open(file,"r")
        prevline = ""
        line = fp.read()

        #the extract regex will look for multiple instance of ddblock
        #pairs, ensuring no ddblock tags are in the main body 

        # FIXME: there's a possiblity of terminating a block extract
        #too early if prefix matching is being used.  Consider nested
        #blocks where the tags both have the searched prefix

        sre = """            
                             (\@ddblock_begin.%s%s             # the first group is the line defined with ddblock_begin
                              (?:[\n]|\s[^\n]*[\n]))         #  using some regex look ahead, we ensure the ddblock tag ends with either a newline 
                                                             #  or white space and that we take in the entire ddblock tag
                             ((?:(?!\@ddblock_end.%s%s).)*)    # the second group is everything between tags that does not contain a ddblock_end
                             (^[^\n]*                        # the third group begins with ddblock_end tag 
                              \@ddblock_end.%s%s)                   
                           """

        if(extract_prefix):
            reblock = re.compile(sre%(tag, "\w*", tag, "\w*", tag, "\w*"),
                                 re.I | re.M | re.S | re.VERBOSE) #multiline, case insensetive
        else:
            reblock = re.compile(sre%(tag, "", tag, "", tag, ""),
                                 re.I | re.M | re.S | re.VERBOSE) #multiline, case insensetive
        m = reblock.findall(line)  # find all the matches of the block tag in question

        if(m):
            # if multiple matches are found we need to loop around them allways pulling out the middle block to be extracted
            for i in range(len(m)):
                verbosePrint("")
                verbosePrint("--- Text prior to the block ---")
                verbosePrint(m[i])
                verbosePrint("--- The block ---")
            #always print group 2 out to the screen, as this is the extracted block
                print m[i][1],
                verbosePrint("--- Text after the block ---")
                verbosePrint(m[i])
        else:
            verbosePrint("No match with block regex")

            
################### END extractBlock method ##########################

###################### updateBlock method ############################
#define the updateBlock method which takes one file and modifies it
#CURRENT LIMITATION: update block only replaces one block per file
def updateBlock(tag, file, blockDict):
    #open the pipe from dxext direct stderr in too to make sure we
    #see it to report a problem if it's not there
    ext = os.popen("dxext.py "+file+" 2>&1").readline().strip()
    if("not found" in ext):
        print "ERROR: dxext.py must be in the path - blocks not updated"
        print "(not found in ", os.environ["PATH"], " )"
        exit(1)

    #ignore some files, like starting with ".", being in a directory
    #starting with "." (like .cvs or .svn), or a file ending in "~" or
    #.bak.  To implement this the regexs are geared to match valid
    #files.

    reslashdot = re.compile("/[^\.]")
    redot      = re.compile("^[^\.]")
    retilda    = re.compile("[^\~]$")
    rebak      = re.compile("(?!\.bak).{4}$")
    if((reslashdot.search(file) or redot.search(file)) 
       and retilda.search(file) and (rebak.search(file) or len(file)<4) ):
        if(len(ext)==0):
            ext = "sh"
        print file,"-",
        if(blockDict.has_key(ext)==False):
            print "no block file for extension", ext
            return 
        else:
            verbosePrint("Found dictionary entry for " + ext + " file type")

        assert os.path.isdir(file)==False

        fp = open(file,"r")
        prevline = ""
        line = fp.read()

        #we want the line starting ddblock 
        sre = """        # this is a multiline match (many lines 
                                       #    in a single string)
                           \A(.*       # the first group is whatever is prior to 
                                       #   the block (\A gives us the start of string)
                               ^[^\n]* # the first group ends with the match from the line after ddblock 
                               @ddblock_begin.%s[^\n]*[\n])
                             (.*)      # the second group is block between the ddblock_begin to ddblock_end 
                             (^[^\n]*  # the third group begins with ddblock_end tag and goes to the end of the file
                               @ddblock_end.%s[^\n]*[\n]
                               .*)\Z                   
                           """

        reblock = re.compile(sre%(tag, tag),
                           re.I | re.M | re.S | re.VERBOSE) #multiline, case insensetive
        m = reblock.search(line)

        if(m):
            verbosePrint("")
            verbosePrint("--- Text prior to the block ---")
            verbosePrint(m.group(1))
            verbosePrint("--- The block ---")
            verbosePrint(m.group(2))
            verbosePrint("--- Text after the block ---")
            verbosePrint(m.group(3))
        else:
            verbosePrint("No match with block regex")


        #if we found a block
        if(m):
            verbosePrint("  Found a ddblock " + tag) 

            #######  make a backup of the existing file #######
            bakfilename = file+".bak"
            # since we're saving a backup to a temp directory, strip
            # off any "../" in relative paths
            redotdotslash = re.compile("^\.\./")
            mdds = redotdotslash.match(bakfilename)
            if(mdds):
                bakfilename = bakfilename.lstrip("./")
                
            #if there is a non-relativized (but not absolute)
            #path make a directory for the back up file if it
            #doesn't exist
            if("/" in bakfilename):
                bakfiledir = bakfilename.rsplit("/",1)[0]
                if not os.path.isdir(tmppath+"/"+bakfiledir):
                    os.makedirs(tmppath+"/"+bakfiledir)

            #now the tmppath can be added
            bakfilename = tmppath+"/"+bakfilename
            bakfp = open(bakfilename, "w")
            bakfp.write(line)
            ##############  done making a backup #############

            #and now overwrite original file with updated block
            fp.close()
            fp = open(file,"w")
            
            #WLP - removed unnecessary newline
            fp.write(m.group(1)+blockDict[ext]+m.group(3))
            fp.close()
            print "ddblock \"" + tag + "\" was updated!"

        else:
            #otherwise we are not inserting a tag
            print "did not find matching ddblock %s - no update" % tag,
            fp.close()
    else:
        verbosePrint("detected a file type that should not be printed: " + file)
        if(reslashdot.search(file)): x=1
        else: verbosePrint("  Has a leading slash.")
        if(redot.search(file)): x=1
        else: verbosePrint("  Has a leading dot.")
        if(retilda.search(file)): x=1
        else: verbosePrint("  Has a trailing tilda.")
        if(rebak.search(file)): x=1
        else: 
            if(len(file)>3): 
                verbosePrint("  Has a trailing \"bak\".")


################### END updateBlock method ##########################



#back to the mainline program...

#For replacing mode, the block text is collected (non-recursively)
#from all matching files in the specified directory (UXBLOCKPATH)
#organizing them in a dictionary indexed by extension.
#
#  We assume that a single block for a particular tag will exist in
#  file labeled "tag-ext.txt", where "tag" is the identifier for the
#  block which is passed in as an argument for this script, and "ext"
#  is the extension to which this block may be applied to.  This
#  allows block writers to include the appropriate comment formatting
#  for that file type.  Additionally, a text file with no extension is
#  assumed to be a shell script, so when replacing blocks of an
#  extensionless file, the script will pull from the tag-sh.txt.  If
#  an extension is found that does not appear in this directory, the
#  blocks of that file are not updated.  Tags are considered unique,
#  so file name conflicts should be minimal, but designers may collect
#  different block text for the same tags by locating them in
#  different directories (e.g. liberal copyright vs. a strict
#  copyright).


from os.path import join

for tag in tags:
    if("replace" in mode):
        blocks = {}
        reblock = re.compile("%s-(.*).txt$" % tag)
        for file in os.listdir(blockpath):
            m = reblock.match(file) #the match is of the extension in group 1
            if m and os.path.isdir(file)==False:
                verbosePrint("Adding block from file " + file + " to dictionary")
                fp = open(join(blockpath,file), "r")
# WLP - removing this rstrip causes the source file to keep all trailing new lines
#               blocks[m.group(1)] = fp.read().rstrip()
                blocks[m.group(1)] = fp.read()

        if(len(blocks)<1):
            print "WARNING: No block files found! Expected files of the form %s-(extension).txt"%tag

    if(os.path.isdir(target)==False):
        verbosePrint("Checking file: " + target)
        if("replace" in mode):
            if(len(blocks)>0):
                updateBlock(tag, target, blocks)
        else:
            extractBlock(tag, target, extract_prefix)
    else:
        for root, dirs, files in os.walk(target):
            for file in files:
                verbosePrint("Processing file " + join(root,file))
                if("replace" in mode):
                    if(len(blocks)>0):
                        updateBlock(tag, join(root,file),blocks) 
                else:
                    extractBlock(tag, join(root,file), extract_prefix) 


#call updateBlock on single file or recursively over the directory structure


            

